"""
addons/eventscripts/_libs/python/smauth.py
Version 1.02.002

Licensed under:
http://creativecommons.org/licenses/by-nc-nd/3.0/

If wish to alter the file/contents or re-use some parts please contact me.

(C) [#OMEGA] - K2 2009-2011
"""

'''
Dev Notes:

Add check for existence of events provided in smauth.ini
'''

# ##### #
# Psyco #
# ##### #

import psyco
psyco.full()

# #################### #
# Eventscripts Imports #
# #################### #

import es
import keyvalues
import gamethread

# Custom ES Imports

import pubcvarlib

# ############## #
# Python Imports #
# ############## #

from ConfigParser import SafeConfigParser
from os.path import exists

# ########### #
# Script Info #
# ########### #

info = es.AddonInfo()
info.name = "Sourcemod Python Authprovider"
info.version = "1.02.002"
info.author = "[#OMEGA] - K2"
info.url = "http://www.german-slaughterhouse.de"
info.description = "Provides auth access to sm files."
info.tags = "sm sourcemod auth provider"


if pubcvarlib.Exists('es_smauth_py_ver'):
    pubcvarlib.Unregister('es_smauth_py_ver')
pubcvarlib.Register('es_smauth_py_ver', info.version)

# ############ #
# Global stuff #
# ############ #

selfmodule = __import__(__name__)  

gamedir = es.getString('eventscripts_gamedir')

config = SafeConfigParser()
if exists(gamedir + '/cfg/smauth/smauth.ini'):
    config.read(gamedir + '/cfg/smauth/smauth.ini')
else:
    raise IOError, 'smauth.ini was not found!'

# ####### #
# Classes #
# ####### #

#selfmodule = __import__(__name__)  
#def unregisterForEvent(self, listener, eventname):
#  def registerForEvent(self, listener, eventname, callback):

class InputError(Exception):
    pass

class __Admin(object):
    """
    Using this class (or the object of this class) should be avoided unless you really need it.
    """

    def __init__(self):
        self.__Parse()
        
    # ####### #
    # Private #
    # ####### #
        
    def __Parse(self):
        self.users = []
        self.groups = {}
        self.levels = {}
        # Since there is a bug with multiline comments in the keyvlaue parser, the multiline comments need to be removed
        # As I don't want to overwrite any of the default admin files, I'll make a work copy.
        for file in ['admin_groups.cfg','admin_levels.cfg','admins.cfg']:
            original = open(gamedir + '/addons/sourcemod/configs/' + file)
            content = original.read()
            # Loop though until all multiline comments are removed
            while content.find('/*') != -1:
                content = content[0:content.find('/*')] + content[content.find('*/') + 2:]
            copy = open(gamedir + '/addons/eventscripts/_libs/python/smauth/' + file,'w')
            copy.write(content)
            # Close both files
            copy.close()
            original.close()
        # Open the copies with keyvalues
        admin_groups = keyvalues.KeyValues(filename = gamedir + '/addons/eventscripts/_libs/python/smauth/admin_groups.cfg')
        admin_levels = keyvalues.KeyValues(filename = gamedir + '/addons/eventscripts/_libs/python/smauth/admin_levels.cfg')
        admins = keyvalues.KeyValues(filename = gamedir + '/addons/eventscripts/_libs/python/smauth/admins.cfg')
        
        # ################# #
        # admins_simple.ini #
        # ################# #
        adminflatfile = open(gamedir + '/addons/sourcemod/configs/' + 'admins_simple.ini') 
        lines = adminflatfile.read().split('\n')
        for line in lines:
            # Ingore comments and blank lines
            if not line.startswith('//') and len(line.strip()):
                # Prepare variables
                auth,id,options = None,None,{'name': None, 'flags': '', 'password': None, 'group': [], 'immunity': 0}
                #Break down the line
                adminline = []
                for aline in line.split('"'):
                    aline = aline.replace('\t','')
                    if len(aline) and not aline.startswith('//'):
                        adminline.append(aline)
                # Parse the auth types
                if adminline[0].startswith('STEAM_'):
                    auth, id = 'steam',adminline[0]
                elif adminline[0][0] == '!':
                    auth, id = 'ip',adminline[0][1:]
                else:
                    auth, id = 'name',adminline[0]
                # Parse admin flags
                for item in adminline[1].split(':'):
                    if item.isdigit():
                        options['immunity'] = item
                    elif item[0] == '@':
                        options['group'] = item[1:]
                    else:
                        options['flags'] = item
                # Parse the password (incase it exists)
                if len(adminline) == 3:
                    options['password'] = adminline[2]
                # Finally add the admin to the list
                self.AddUser(auth,id,options)
            # Close the file as we finished up
            adminflatfile.close()
        
        # ########## #
        # admins.ini #
        # ########## #
        for entry in admins:
            options = {'name': None, 'flags': '', 'password': None, 'group': [], 'immunity': 0}
            options['name'] = entry.getName()
            # Loop though the keyvalue pairs
            for item in entry:
                itemname = item.getName()
                if itemname == 'group':
                    options['group'].append(item.getString())
                elif itemname not in ['auth','identity']:
                    options[itemname] = item.getString()
            # Add the User
            self.AddUser(entry['auth'],entry['identity'],options)
        
        # ################ #
        # admin_groups.ini #
        # ################ #
        for entry in admin_groups:
            self.AddGroup(entry.getName(),entry['flags'] if 'flags' in entry.keys() else '', entry['immunity'] if 'immunity' in entry.keys() else 0)
        
        # ################ #
        # admin_levels.ini #
        # ################ #
        for entry in admin_levels['Flags']:
            self.levels[entry.getName()] = entry.getString()
    def __GetInt(self,var):
        if type(var) in (int,float,long):
            return var
        elif type(var) == str and var.isdigit():
            return int(var)
        elif var == None:
            return 0
        else:
            raise InputError, '%s can\'t be parsed to an integer value.' % var
    def __AddFlags(self,flags1,flags2):
        """
        Adds 2 flag strings togehter.
        """
        addedflags = ''
        for flag in flags1 + flags2:
            if flag not in addedflags:
                addedflags += flag
        return addedflags
    # ###### #
    # Public #
    # ###### #
    def AddUser(self, auth='steam', identity=None, options = {'name': None, 'flags': '', 'password': None, 'group': [], 'immunity': 0}):
        # A bit of error checking
        if not auth:
            raise InputError, 'Auth type (steam,ip,name) must be provided!'
        elif auth not in ('steam','ip','name'):
            raise InputError, 'Auth type must be either steam, ip or name.\nAuth Type was: %s' % auth
        if not identity:
            raise InputError, 'Identity value must be provided!'
        # Make sure immunity is set to a proper value
        if options.has_key('immunity'):
            options['immunity'] = self.__GetInt(options['immunity'])
        # Same for steamid to ensure compability between old and new steam auth metod
        if auth == 'steam':
            # Remove the STEAM_X: part from the steamid
            identity = identity[8:]
        
        # Parsing for the group
        if options.has_key('group') and type(options['group']) == str:
            options['group'] = list(options['group'])
        
        # Add the user
        temp = {'auth': auth,'identity': identity}
        temp.update(options)
        self.users.append(temp)
    def AddGroup(self, name=None, flags='', immunity=0):
        # Error Checking
        if not name:
            raise InputError, 'Group name must be defined'
        # Check for duplicates
        if self.groups.has_key(name):
            raise InputError, 'This group already exists. Groups must be unique!'
        # Add the group
        self.groups[name] = {'flags': flags, 'immunity': self.__GetInt(immunity)}
    def GetUser(self,identity,auth='steam'):
        """
        Gets the user object with parsed permissions.
        If the user is not in there, it returns False.
        """
        for user in self.users:
            if user['auth'] == auth and user['identity'] == identity:
                # Adds up immunity and flags for each group the user is in
                for admingroup in ['Default',] + user['group']:
                    for group in self.groups:
                        if group == admingroup:
                            user['immunity'] = self.groups[group]['immunity'] if self.groups[group]['immunity'] > user['immunity'] else user['immunity']
                            user['flags'] = self.__AddFlags(user['flags'],self.groups[group]['flags'])
                return user
        return False
    def Parse(self):
        self.__Parse()
        
class PlayerAccess(object):
    """
    The purpose of this class is to simplify the access to the data parsed by the libary. It is possible to use the Admin object directly, however, unless it's not avoidable this class should be preferred.
    """
    def __init__(self, userid, authtype='steam', authinfo=None):
        """
        Creates a new player object for the specified userid.
        
        @userid        - The userid of the player
        @authtype    - The authtype that is used when checking permissions. Can be either steam, ip or name (Default: steam)
        @authinfo   - Optional argument that represenets the authentification tag depending on authtype, can be one of the following:
                      - SteamID of player (for authtype 'steam')
                      - IP Address of player (for authtype 'ip')
                      - Name of player (for authtype 'name')
        
        Raises InputError if the type is invalid (not steam, ip or name) 
        """
        self.userid = userid
        self.authtype = authtype.lower()
        if authtype == 'steam':
            if authinfo:
                if authinfo.startswith('STEAM_'):
                    authinfo = authinfo[8:]
                self.authinfo = authinfo
            else:
                self.authinfo = es.getplayersteamid(userid)[8:] 
        elif authtype == 'ip':
            self.authinfo = authinfo if authinfo else es.createplayerlist(userid)[str(userid)]['address'].split(':')[0]
        elif authtype == 'name':
            self.authinfo = authinfo if authinfo else es.getplayername(userid)
        else:
            raise InputError, 'Auth type must be either steam, ip or name.\nAuth Type was: %s' % auth
        # Get the userinfo
        self.user = Admin.GetUser(self.authinfo,self.authtype)
    def IsUser(self):
        """
        Checks whether the user is stored in the database
        
        Returns True if the user with the specified auth is stored in admin files
        Returns False otherwise 
        """
        return bool(self.user)
    def IsAdmin(self):
        """
        Checks whether the user has basic admin access (b) or not.
        
        Returns True if the user has the basic admin (b) flag
        Returns False if the user does not exist or if the flag can not be found.
        """
        if self.IsUser():
            for flag in 'bz':
                if flag in self.user['flags']:
                    return True
        return False
    def IsRoot(self):
        """
        Checks whether the user has root admin access (z) or not.
        
        Returns True if the user has root access.
        Returns False otherwise.
        """
        return self.HasFlag()
    def HasFlag(self, adminflag=''):
        """
        This functions checks whether the user has the specified flag. It also checks against whether he has root access and thus is able to use the function without having the flag itself.
        
        @adminflag    - A single character which represents the admin flag.
        
        Returns True if the user has the specified flag or has root access.
        Returns False if the user does not exist or if the flag can not be found.
        
        Raises TypeError if adminflag is not a string.
        """
        if not isinstance(adminflag, basestring):
            raise TypeError, "Invalid type for adminflags specified. Type received was %s." % type(adminflags)
        
        if self.IsUser():
            for flag in adminflag + 'z':
                if flag in self.user['flags']:
                    return True
        return False
    def HasFlags(self, adminflags=()):
        """
        This functions checks whether the user has all specified flag. It also checks against whether he has root access and thus is able to use the function without having the flag itself.
        
        @adminflags    - Any iterable type (dict, list, tuple or strings)
        
        Returns True if the user has all the specified flags (or root flag).
        Returns False if the has one of the specified flags or if the flag can not be found.
        
        Raises TypeError if the type is invalid.
        """
        if not isinstance(adminflags, (dict, list, tuple, basestring)):
            raise TypeError, "Invalid type for adminflags specified. Type received was %s." % type(adminflags)
        
        if not self.IsUser():
            return False
            
        if 'z' in self.user['flags']:
            return True
            
        for flag in adminflags:
            if flag not in self.user['flags']:
                return False
        
        return True
    def GetImmunity(self):
        """
        This function is used to receive the immunity level of the user. It will always return the highest applicable immnunity.
        
        Returns the immunity value of th euser
        Returns False otherwise
        """
        return self.user['immunity'] if self.IsUser() else False
    def InGroup(self,group):
        """
        This functions checks whether the user is in the specifed admin group. This functino is case-senstive.
        
        @group    - The admin group
        
        Returns True if the user is in the specified group.
        Returns False if the user does not exist or if the group can not be found.
        """
        return True if self.IsUser() and group in self.user['group'] else False
    def CanTarget(self,userid):
        """
        This functions checks whether the user can target the player with the specified userid with an admin command. Depending on sm_immunity_mode the according value may vary. See SourceMod documentation for more info on what sm_immunity_mode changes.
        
        @userid    - The userid of the targeted player
        
        Returns True if the user can target the specified userid with an admin command.
        Returns False if the user does not exist or if the userid can not be targeted by the user.
        """
        if not self.IsUser():
            return False
        immunity_mode = es.getInt('sm_immunity_mode')
        # If 0 immunity levels are ingored anyway...
        if immunity_mode:
            # Root admins can target anyone, no matter the immunity.
            if not self.IsRoot():
                for authtype in ('steam','ip','name'):
                    target = PlayerAccess(userid,authtype)
                    if target.IsUser():
                        if immunity_mode == 1:
                            return self.GetImmunity() >= target.GetImmunity()
                        elif immunity_mode == 2:
                            return self.GetImmunity() > target.GetImmunity()
                        elif immunity_mode == 3:
                            if self.GetImmunity() == 0 and target.GetImmunity() == 0:
                                return True
                            else:
                                return self.GetImmunity() > target.GetImmunity()
        return True

# ######### #
# Functions #
# ######### #

# ###### #
# Pivate #
# ###### #
def __RegisterAutoRefresher():
    event = config.get('general','event')
    if len(event):
        es.addons.registerForEvent(selfmodule, event, __AutoRefreshEvent)

def __AutoRefreshEvent(ev):
    Admin.Parse()

# ###### #
# Public #
# ###### #
def FlagToLevel(flag):
    """
    Converts the flag to the according sm access level.
    
    @flag    - the flag that is supposed to be converted to the coressponding level
    
    Raises InputError if the flag doesn't exist.
    """
    for level in Admin.levels:
        if Admin.levels[level] == flag:
            return level
    raise InputError, 'Flag %s does not exist!' % flag

def LevelToFlag(level):
    """
    Converts the sm accesslevel to the according sm flag.
    This is case-sensitive.
    
    @level    - the name of the admin level that is supposed to be converted to the coressponding flag
    
    Raises InputError if the level doesn't exist.
    """
    if IsValidLevel(level):
        return Admin.levels[level]
    else:
        raise InputError, 'Access level %s does not exist' % level

def IsValidFlag(flag):
    """
    This function can be used to check whether the specified flag is valid. This function is case-sensitive.
    
    @flag    - the flag you want to validate
    
    Returns True if the specified flag is valid and exists.
    Returns False otherwise 
    """
    for level in Admin.levels:
        if Admin.levels[level] == flag:
            return True
    return False
    
def IsValidLevel(level):
    """
    This function can be used to check whether the specified admin level is valid. This function is case-sensitive.
    
    @level    - the level you want to validate
    
    Returns True if the specified level is valid and exists.
    Returns False otherwise
    """
    return Admin.levels.has_key(level)
    
def IsValidGroup(group):
    """
    This function can be used to check whether the specified admin group is valid. This function is case-sensitive.
    
    @group    - the name of the group you want to validate
    
    Returns True if the specified group is valid and exists.
    Returns False otherwise
    """
    return group in Admin.groups
#
#  
#
Admin = __Admin()
__RegisterAutoRefresher()